(declaim (optimize (speed 1) (safety 3) (debug 3)))

(ql:quickload "osicat")
(ql:quickload "cl-fad")
(ql:quickload "cl-heredoc")

(defvar *produced-output*        "output.txt")
(defvar *solution-output*        "main")
(defvar *bash*                   "/bin/bash")

(defvar *text-compare-helper*    "compare-text.sh")
(defvar *run-a-test-helper*      "run-a-test.sh")
(defvar *cleanup-all-program*    "cleanup-all.sh")

(defvar *string-if-matched*      "Passed!")
(defvar *string-if-matched-not*  "*** Failed! ***")

(defvar *problems-list*          nil)
(defvar *problems-read*          nil)
(defvar *tests-list*             nil)
(defvar *tests-read*             nil)

(defparameter *working-directory* (namestring (osicat:current-directory)))

(defparameter *config-class-structure*
  `((problem-code "") (problem-name "")
    (from-test    1)  (to-test      1)
    (input-directory  #p"test/")
    (solution-file    "test/printaline.go")
    (solution-output  "jail/output.txt")
    (language     "go")
    (time-limit   2.0)
    (mem-limit    (* 2.0 1024 1024))))

(defparameter *problem-class-structure*
  `((problem-code "") (problem-name "")
    (input-directory  #p"test/")
    (time-limit   2.0)
    (mem-limit    (* 2.0 1024 1024))))

(defparameter *solution-class-structure*
  `((problem-code  "")
    (solution-file   "test/printaline.go")
    (solution-output "jail/output.txt")
    (language     "go")))

(defparameter *test-class-structure*
  `((solution-file   "test/printaline.go")
    (solution-program "jail/main")
    (solution-output "jail/output.txt")
    (language     "go")
    (problem-code    "")
    (from-test       1)
    (to-test         1)))

(defun make-accessor (slot-name)
  (intern (concatenate 'string
                       "CONF-"
                       (string-upcase (string slot-name)))))

(defun make-initarg (slot-name)
  (intern (string slot-name) "KEYWORD"))

(defun make-class-from-structure (s
                                  &optional (class-name 'judge-config))
  (eval `(defclass ,class-name ()
           ,(loop
               for slot in s
               collect
               (let* ((slot-name (nth 0 slot))
                      (slot-init-value (nth 1 slot)))
                 (list slot-name
                       :accessor (make-accessor slot-name)
                       :initarg  (make-initarg slot-name)
                       :initform slot-init-value))))))

(defun make-dump-string (class-structure)
  (format nil "~{~{~a~*~}: ~~15a~~%~}" class-structure))

(defun get-slot-value (slot-name obj)
  (eval `(funcall #',(intern (concatenate 'string "CONF-"
                                          (symbol-name slot-name))) ,obj)))

(defun make-dump-parameters (class-structure class-object)
  (loop
     for slot in class-structure
     collect (get-slot-value (first slot) class-object)))

(defun dump-class (obj class-structure &optional (stream *standard-output*))
  (apply #'format stream (make-dump-string class-structure)
         (make-dump-parameters class-structure obj)))

(defun replace-extension (file-name old-ext new-ext)
  "Replace a file extension."
  (concatenate 'string
               (subseq file-name 0 (search old-ext file-name :from-end t))
               new-ext))

(defun cleanup-all ()
  "Clean-up all temporary files."
  (run-program *bash*
               (list *cleanup-all-program*)
               :wait t :output nil))

(defun generate-compile-command (&key (language "go")
                                 (working-directory *working-directory*))
  (concatenate 'string working-directory "compile-" language ".sh"))

(defun compile-solution (language solution-file
                         &key (solution-output "jail/main")
                         (working-directory *working-directory*))
  "Compile a source file."

  ;; Cleaning up first
  (cleanup-all)

  ;; Then compile
  (let* ((out-stream (make-string-output-stream))
         (err-stream (make-string-output-stream))
         (p (run-program *bash*
                         (list (generate-compile-command
                                :language language
                                :working-directory working-directory)
                               solution-file solution-output)
                         :wait t :input nil :pty nil
                         :output out-stream
                         :error err-stream)))
    (values (process-exit-code p)
            (get-output-stream-string out-stream)
            (get-output-stream-string err-stream))))

(defun compare-text (file-1 file-2 &key
                     (text-compare-helper *text-compare-helper*))
  "Comparing file-1 and file-2 using a helper and text comparison method."
  (process-exit-code (run-program *bash*
                                  (list text-compare-helper
                                        (namestring file-1)
                                        (namestring file-2))
                                  :wait t :input nil :output nil)))

(defun make-test-file (file-type number)
  "Make an input/output/answer file for testing."
  (if (stringp number)
      (concatenate 'string number "." file-type)
      (concatenate 'string (write-to-string number) "." file-type)))

(defun run-a-test (&key
                   (helper            *run-a-test-helper*)
                   (solution-program  "jail/main")
                   (test-number       "1")
                   (input-directory   "")
                   (output-file       ""))
  "Run a test case using helper run-a-test.sh"
  (let* ((out-stream (make-string-output-stream))
         (err-stream (make-string-output-stream))
         (p (run-program *bash*
                        (list helper (namestring solution-program)
                              test-number
                              (namestring input-directory)
                              (namestring output-file))
                        :output out-stream
                        :output err-stream)))
    (values (process-exit-code p)
            (get-output-stream-string out-stream)
            (get-output-stream-string err-stream))))

(defun run-a-test-set (&key
                       (helper *run-a-test-helper*)
                       test-obj problem-obj
                       (checking-method   (function compare-text)))
  "Run a test set."
  (loop
     for test-num from (conf-from-test test-obj) to (conf-to-test test-obj)
     collect
       (progn
         (run-a-test :helper           helper
                     :solution-program (conf-solution-program test-obj)
                     :test-number      (format nil "~a" test-num)
                     :input-directory  (conf-input-directory problem-obj)
                     :output-file      (conf-solution-output test-obj))
         (list test-num
               (funcall checking-method (conf-solution-output test-obj)
                        (make-test-file "ans" test-num))))))

(defun print-test-set-result (results &key (stream *standard-output*))
  "Print the result for each test in a test set."
  (format stream "~{~{~5d ~a~}~%~}" results))

(defun get-compile-info (process)
  (values
   (process-exit-code process)
   (get-output-stream-string (process-output process))
   (get-output-stream-string (process-error  process))))

(defun initialize (&key
                   (problem-conf-file "conf-problems.lisp")
                   (test-conf-file    "conf-tests.lisp"))
  "Initialize the system, run only once."
  (make-class-from-structure *config-class-structure*)
  (make-class-from-structure *problem-class-structure*  'judge-problem)
  (make-class-from-structure *solution-class-structure* 'judge-solution)
  (make-class-from-structure *test-class-structure*     'judge-test)
  (load (compile-file problem-conf-file))
  (load (compile-file test-conf-file))
  (setf *problems-list* *problems-read*
        *tests-list*    *tests-read*))
